/*
 *  tslib/plugins/tcc-raw.c
 *
 *  Original version:
 *  Copyright (C) 2001 Russell King.
 *
 *  Rewritten for the Linux input device API:
 *  Copyright (C) 2002 Nicolas Pitre
 *
 * This file is placed under the LGPL.  Please see the file
 * COPYING for more details.
 *
 *
 * Read raw pressure, x, y, and timestamp from a touchscreen device.
 */
#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <limits.h>

#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/time.h>
#include <sys/types.h>

#include <linux/input.h>
#ifndef EV_SYN /* 2.4 kernel headers */
# define EV_SYN 0x00
#endif
#ifndef EV_ABS /* 2.4 kernel headers */
#define EV_ABS 0x03
#endif
#ifndef EV_CNT
# define EV_CNT (EV_MAX+1)
#endif
#ifndef ABS_CNT
# define ABS_CNT (ABS_MAX+1)
#endif
#ifndef KEY_CNT
# define KEY_CNT (KEY_MAX+1)
#endif

#ifndef ABS_MT_POSITION_X
# define ABS_MT_POSITION_X       0x35    /* Center X ellipse position */
# define ABS_MT_POSITION_Y       0x36    /* Center Y ellipse position */
#endif

#include "tslib-private.h"

//DEBUG-------------------
#if 0
#define tccdprint printf
#else
#define tccdprint
#endif

#define GRAB_EVENTS_WANTED	1
#define GRAB_EVENTS_ACTIVE	2

/* includes available in 2.6.30-rc5 */
#ifndef BTN_TOOL_QUADTAP
#define BTN_TOOL_QUADTAP	0x14f	/* Four fingers on trackpad */
#define ABS_MT_TOUCH_MAJOR	0x30	/* Major axis of touching ellipse */
#define ABS_MT_TOUCH_MINOR	0x31	/* Minor axis (omit if circular) */
#define ABS_MT_WIDTH_MAJOR	0x32	/* Major axis of approaching ellipse */
#define ABS_MT_WIDTH_MINOR	0x33	/* Minor axis (omit if circular) */
#define ABS_MT_ORIENTATION	0x34	/* Ellipse orientation */
#define ABS_MT_POSITION_X	0x35	/* Center X ellipse position */
#define ABS_MT_POSITION_Y	0x36	/* Center Y ellipse position */
#define ABS_MT_TOOL_TYPE	0x37	/* Type of touching device */
#define ABS_MT_BLOB_ID		0x38	/* Group a set of packets as a blob */
#define ABS_MT_TRACKING_ID	0x39	/* Unique ID of initiated contact */
#define SYN_MT_REPORT		2
#define MT_TOOL_FINGER		0
#define MT_TOOL_PEN		1
#endif

/* includes available in 2.6.33 */
#ifndef ABS_MT_PRESSURE
#define ABS_MT_PRESSURE		0x3a	/* Pressure on contact area */
#endif


/* includes available in 2.6.36 */
#ifndef ABS_MT_SLOT
#define ABS_MT_SLOT		0x2f	/* MT slot being modified */
#endif

/* includes available in 2.6.38 */
#ifndef ABS_MT_DISTANCE
#define ABS_MT_DISTANCE		0x3b	/* Contact hover distance */
#endif


#define SLOT_MAX (10)
#define TSLIB_SUPPORT_MULTI	0

typedef enum {
	EMVARIANCE_FAKE_PRE=0,
	EMVARIANCE_FAKE_NEED,
	EMVARIANCE_FAKE_DONE
} EMVARIANCE_FAKE;

struct tslib_input {
	struct tslib_module_info module;

	int	current_x[SLOT_MAX];
	int	current_y[SLOT_MAX];
	int	current_p[SLOT_MAX];
	int	current_v[SLOT_MAX];	// validate..
	int current_pen_up[SLOT_MAX];
	int	current_slot;

	
	int	sane_fd;
	int	using_syn;
	int	grab_events;
	int multi_touch;
	EMVARIANCE_FAKE multi_touch_need_fake;	// fake to variance noise check..!
};

#define DIV_ROUND_UP(n,d) 		(((n) + (d) - 1) / (d))
#define BIT(nr)                 (1UL << (nr))
#define BIT_MASK(nr)            (1UL << ((nr) % BITS_PER_LONG))
#define BIT_WORD(nr)            ((nr) / BITS_PER_LONG)
#define BITS_PER_BYTE           8
#define BITS_PER_LONG           (sizeof(long) * BITS_PER_BYTE)
#define BITS_TO_LONGS(nr)       DIV_ROUND_UP(nr, BITS_PER_BYTE * sizeof(long))

static int test_bit(int nr, const unsigned long *addr)
{
	return 1UL & (addr[BIT_WORD(nr)] >> (nr & (BITS_PER_LONG-1)));
}

static int check_fd(struct tslib_input *i)
{
	struct tsdev *ts = i->module.dev;
	int version;
	long evbit[BITS_TO_LONGS(EV_CNT)];
	long absbit[BITS_TO_LONGS(ABS_CNT)];
	long keybit[BITS_TO_LONGS(KEY_CNT)];

	unsigned long absbits[(ABS_MAX+31)/32];

	if(ioctl(ts->fd, EVIOCGBIT(EV_ABS, sizeof(absbits)), absbits) < 0) {
		fprintf(stderr, "tslib: Selected device is not a Linux input event device\n");
		ioctl(ts->fd, EVIOCGRAB, 0);
		return -1;
	}

	if ((0 == test_bit(ABS_MT_POSITION_X, absbits)) || (0 == test_bit(ABS_MT_POSITION_Y, absbits)) || (0 == test_bit(ABS_MT_TOUCH_MAJOR, absbits))) {
		#ifdef DEBUG
		fprintf(stderr, "tslib: could not multi-touch the device\n");
		fprintf(stderr, "ABS_MT_POSITION_X=%d\n", test_bit(ABS_MT_POSITION_X, absbits));
		fprintf(stderr, "ABS_MT_POSITION_Y=%d\n", test_bit(ABS_MT_POSITION_Y, absbits));
		fprintf(stderr, "ABS_MT_TOUCH_MAJOR=%d\n", test_bit(ABS_MT_TOUCH_MAJOR, absbits));
		#endif
		i->multi_touch = 0;
	}
	else
	{
		i->multi_touch= 1;
		i->using_syn = 1;
	}

	tccdprint("tslib: multi_touch=%d\r\n", i->multi_touch);

	if(!i->multi_touch)	{
		if (ioctl(ts->fd, EVIOCGVERSION, &version) < 0) {
			fprintf(stderr, "tslib: Selected device is not a Linux input event device\n");
			return -1;
		}

		if (version < EV_VERSION) {
			fprintf(stderr, "tslib: Selected device uses a different version of the event protocol than tslib was compiled for\n");
			return -1;
		}

		if ( (ioctl(ts->fd, EVIOCGBIT(0, sizeof(evbit)), evbit) < 0) ||
			!(evbit[BIT_WORD(EV_ABS)] & BIT_MASK(EV_ABS)) ||
			!(evbit[BIT_WORD(EV_KEY)] & BIT_MASK(EV_KEY)) ) {
			fprintf(stderr, "tslib: Selected device is not a touchscreen (must support ABS and KEY event types)\n");
			return -1;
		}

		if ((ioctl(ts->fd, EVIOCGBIT(EV_ABS, sizeof(absbit)), absbit)) < 0 ||
			!(absbit[BIT_WORD(ABS_X)] & BIT_MASK(ABS_X)) ||
			!(absbit[BIT_WORD(ABS_Y)] & BIT_MASK(ABS_Y))) {
			fprintf(stderr, "tslib: Selected device is not a touchscreen (must support ABS_X and ABS_Y events)\n");
			return -1;
		}

		/* Since some touchscreens (eg. infrared) physically can't measure pressure,
		the input system doesn't report it on those. Tslib relies on pressure, thus
		we set it to constant 255. It's still controlled by BTN_TOUCH/BTN_LEFT -
		when not touched, the pressure is forced to 0. */

		if (!(absbit[BIT_WORD(ABS_PRESSURE)] & BIT_MASK(ABS_PRESSURE))) {
			int islot;
			for(islot=0;islot < SLOT_MAX;islot++)
				i->current_p[islot] = 255;

			if ((ioctl(ts->fd, EVIOCGBIT(EV_KEY, sizeof(keybit)), keybit) < 0) ||
				!(keybit[BIT_WORD(BTN_TOUCH)] & BIT_MASK(BTN_TOUCH) ||
				  keybit[BIT_WORD(BTN_LEFT)] & BIT_MASK(BTN_LEFT))) {
				fprintf(stderr, "tslib: Selected device is not a touchscreen (must support BTN_TOUCH or BTN_LEFT events)\n");
				return -1;
			}
		}

		if (evbit[BIT_WORD(EV_SYN)] & BIT_MASK(EV_SYN))
			i->using_syn = 1;

		if (i->grab_events == GRAB_EVENTS_WANTED) {
			if (ioctl(ts->fd, EVIOCGRAB, (void *)1)) {
				fprintf(stderr, "tslib: Unable to grab selected input device\n");
				return -1;
			}
			i->grab_events = GRAB_EVENTS_ACTIVE;
		}
	}

	return 0;
}

#if TSLIB_SUPPORT_MULTI
static int push_multisample( struct tslib_input *i, struct input_event *ev, struct ts_sample * samp ) 
{
	int islot, num = 0;
	tccdprint("EV_SYNC: SYN_MT_REPORT\n");
	for (islot=0;islot<SLOT_MAX;islot++)
	{
		if (i->current_v[islot])
		{
			samp[num].x = i->current_x[islot];
			samp[num].y = i->current_y[islot];
			samp[num].pressure = i->current_p[islot];
			samp[num].tv = ev->time;
			tccdprint("         SYN_MT_REPORT %d %d %d\n", samp[num].x, samp[num].y, samp[num].pressure);
			
			num++;

			// clean previous value.
			i->current_v[islot] = 0;
			i->current_x[islot] = 0;
			i->current_y[islot] = 0;
			i->current_p[islot] = 0;
		}
	}
	tccdprint("leave. num=%d\n", num);
	return num;
}
#else
#define push_multisample(i, ev, samp) 0
#endif

static int ts_input_read(struct tslib_module_info *inf,
			 struct ts_sample *samp, int nr)
{
	struct tslib_input *i = (struct tslib_input *)inf;
	struct tsdev *ts = inf->dev;
	struct input_event ev;
	int ret = nr;
	int total = 0;
	int pen_up = 0;

	tccdprint("ts_input_read\r\n");
	
	// first check 
	if (i->sane_fd == -1)
		i->sane_fd = check_fd(i);

	if (i->sane_fd == -1)
		return 0;

	if (i->using_syn) {
		while (total < nr) {
			if(i->multi_touch && i->multi_touch_need_fake == EMVARIANCE_FAKE_NEED) {
				ev.type = EV_SYN;
				ev.code = SYN_MT_REPORT;
				i->multi_touch_need_fake = EMVARIANCE_FAKE_DONE;
			}
			else {
				ret = read(ts->fd, &ev, sizeof(struct input_event));
				if (ret < (int)sizeof(struct input_event)) {
					total = -1;
					break;
				}
			}

			switch (ev.type) {
			case EV_KEY:
				switch (ev.code) {
				case BTN_TOUCH:
				case BTN_LEFT:
					tccdprint("BTN_TOUCH\r\n");
					if (ev.value == 0)
						i->current_pen_up[i->current_slot] = 1;
					break;
				}
				break;
			case EV_SYN:
				if (ev.code == SYN_REPORT || ev.code == SYN_MT_REPORT) {
					if(i->multi_touch) {						
						#if TSLIB_SUPPORT_MULTI
						int push_count;
						push_count = push_multisample(i, &ev, samp);
						samp += push_count;
						total += push_count;
						#else
						// Process one point only
						if(!i->current_slot)
						{
							if(i->current_pen_up[i->current_slot])
							{
								tccdprint("EV_SYNC PENUP\r\n");
								samp->x = 0;
								samp->y = 0;
								samp->pressure = 0;
								samp->tv = ev.time;
								i->current_pen_up[i->current_slot] = 0;
								i->multi_touch_need_fake = EMVARIANCE_FAKE_PRE;
							}
							else
							{
								tccdprint("EV_SYNC PENPUSH\r\n");
								samp->x = i->current_x[i->current_slot];
								samp->y = i->current_y[i->current_slot];
								samp->pressure = i->current_p[i->current_slot];
								i->multi_touch_need_fake = (i->multi_touch_need_fake!=EMVARIANCE_FAKE_DONE)?EMVARIANCE_FAKE_NEED:EMVARIANCE_FAKE_DONE;
							}
							samp->tv = ev.time;
							samp++;
							total++;
						}
						#endif
					}
					else
					{
						/* Fill out a new complete event */
						if (i->current_pen_up[i->current_slot]) {
							samp->x = 0;
							samp->y = 0;
							samp->pressure = 0;
							i->current_pen_up[i->current_slot] = 0;
						} else {
							samp->x = i->current_x[i->current_slot];
							samp->y = i->current_y[i->current_slot];
							samp->pressure = i->current_p[i->current_slot];
						}
						samp->tv = ev.time;
						tccdprint("EV_SYNC: SYN_REPORT %d %d %d\n", samp->x, samp->y, samp->pressure);
						#ifdef DEBUG
						fprintf(stderr, "RAW---------------------> %d %d %d %d.%d\n",
								samp->x, samp->y, samp->pressure, samp->tv.tv_sec,
								samp->tv.tv_usec);
						#endif /* DEBUG */
						samp++;
						total++;
					}
				}
				break;
			case EV_ABS:
				switch (ev.code) {
				case ABS_X:
					i->current_x[i->current_slot] = ev.value;
					i->current_v[i->current_slot] = 1;
					tccdprint("EV_ABS: ABS_X=%d\n", ev.value);
					break;
				case ABS_Y:
					i->current_y[i->current_slot] = ev.value;
					tccdprint("EV_ABS: ABS_Y=%d\n", ev.value);
					break;
				case ABS_PRESSURE:
					i->current_p[i->current_slot] = ev.value;
					tccdprint("EV_ABS: ABS_PRESSURE=%d\n", ev.value);
					break;
					
				case ABS_MT_SLOT:
					if (ev.value >= 0) {
						i->current_slot = ev.value;
						tccdprint("EV_ABS: ABS_MT_SLOT=%d\n", ev.value);
						if (i->current_slot >= SLOT_MAX) {
							//fprintf(stderr, "tslib: slot limit error. SLOT_MAX=%d. current slot=%d\n", SLOT_MAX, i->current_slot);
							i->current_slot = 0;
						}
					}
					break;
					
				case ABS_MT_TOUCH_MAJOR:
					tccdprint("EV_ABS: ABS_MT_TOUCH_MAJOR=%d\n", ev.value);
					i->current_p[i->current_slot] = ev.value;
					break;
					
				case ABS_MT_POSITION_X:
					tccdprint("EV_ABS: ABS_MT_POSITION_X=%d\n", ev.value);
					i->current_x[i->current_slot] = ev.value;
					i->current_v[i->current_slot] = 1;
					break;
					
				case ABS_MT_POSITION_Y:
					tccdprint("EV_ABS: ABS_MT_POSITION_Y=%d\n", ev.value);
					i->current_y[i->current_slot] = ev.value;
					i->current_v[i->current_slot] = 1;
					break;
					
				case ABS_MT_PRESSURE:
					i->current_p[i->current_slot] = ev.value;
					tccdprint("EV_ABS: ABS_MT_PRESSURE=%d\n", ev.value);
					break;

				case ABS_MT_TRACKING_ID:
					if(ev.value < 0)
						i->current_pen_up[i->current_slot] = 1;
					else
						i->current_pen_up[i->current_slot] = 0;						
					tccdprint("EV_ABS: ABS_MT_TRACKING_ID=%d\n", ev.value);
					break;
				}
				break;
			}
		}
		ret = total;
	} else {
		unsigned char *p = (unsigned char *) &ev;
		int len = sizeof(struct input_event);

		while (total < nr) {
			ret = read(ts->fd, p, len);
			if (ret == -1) {
				if (errno == EINTR) {
					continue;
				}
				break;
			}

			if (ret < (int)sizeof(struct input_event)) {
				/* short read
				 * restart read to get the rest of the event
				 */
				p += ret;
				len -= ret;
				continue;
			}
			/* successful read of a whole event */

			if (ev.type == EV_ABS) {
				switch (ev.code) {
				case ABS_X:
					if (ev.value != 0) {
						samp->x = i->current_x[i->current_slot] = ev.value;
						samp->y = i->current_y[i->current_slot];
						samp->pressure = i->current_p[i->current_slot];
					} else {
						fprintf(stderr, "tslib: dropped x = 0\n");
						continue;
					}
					break;
				case ABS_Y:
					if (ev.value != 0) {
						samp->x = i->current_x[i->current_slot];
						samp->y = i->current_y[i->current_slot] = ev.value;
						samp->pressure = i->current_p[i->current_slot];
					} else {
						fprintf(stderr, "tslib: dropped y = 0\n");
						continue;
					}
					break;
				case ABS_PRESSURE:
					samp->x = i->current_x[i->current_slot];
					samp->y = i->current_y[i->current_slot];
					samp->pressure = i->current_p[i->current_slot] = ev.value;
					break;
				case ABS_MT_SLOT:
					if (ev.value>=0)
					{
						i->current_slot = ev.value;
						if (i->current_slot>=SLOT_MAX)
						{
							//fprintf(stderr, "tslib: slot limit error. SLOT_MAX=%d. current slot=%d\n",SLOT_MAX,  i->current_slot);
							i->current_slot = 0;
						}
					}
					break;
				case ABS_MT_TOUCH_MAJOR:
					i->current_p[i->current_slot] = ev.value;
					break;
				case ABS_MT_POSITION_X:
					i->current_x[i->current_slot] = ev.value;
					break;
				case ABS_MT_POSITION_Y:
					i->current_y[i->current_slot] = ev.value;
 					break;

				}
				samp->tv = ev.time;
	#ifdef DEBUG
				fprintf(stderr, "RAW---------------------------> %d %d %d\n",
					samp->x, samp->y, samp->pressure);
	#endif /* DEBUG */
				samp++;
				total++;
			} else if (ev.type == EV_KEY) {
				switch (ev.code) {
				case BTN_TOUCH:
				case BTN_LEFT:
					if (ev.value == 0) {
						/* pen up */
						samp->x = 0;
						samp->y = 0;
						samp->pressure = 0;
						samp->tv = ev.time;
						samp++;
						total++;
					}
					break;
				}
			} else {
				fprintf(stderr, "tslib: Unknown event type %d\n", ev.type);
			}
			p = (unsigned char *) &ev;
		}
		ret = total;
	}

	tccdprint("RET=%d\r\n", ret);
	return ret;
}

static int ts_input_fini(struct tslib_module_info *inf)
{
	struct tslib_input *i = (struct tslib_input *)inf;
	struct tsdev *ts = inf->dev;

	if (i->grab_events == GRAB_EVENTS_ACTIVE) {
		if (ioctl(ts->fd, EVIOCGRAB, (void *)0)) {
			fprintf(stderr, "tslib: Unable to un-grab selected input device\n");
		}
	}

	free(inf);
	return 0;
}

static const struct tslib_ops __ts_input_ops = {
	.read	= ts_input_read,
	.fini	= ts_input_fini,
};

static int parse_raw_grab(struct tslib_module_info *inf, char *str, void *data)
{
	struct tslib_input *i = (struct tslib_input *)inf;
	unsigned long v;
	int err = errno;

	v = strtoul(str, NULL, 0);

	if (v == ULONG_MAX && errno == ERANGE)
		return -1;

	errno = err;
	switch ((int)data) {
	case 1:
		if (v)
			i->grab_events = GRAB_EVENTS_WANTED;
		break;
	default:
		return -1;
	}
	return 0;
}

static const struct tslib_vars raw_vars[] =
{
	{ "grab_events", (void *)1, parse_raw_grab },
};

#define NR_VARS (sizeof(raw_vars) / sizeof(raw_vars[0]))

TSAPI struct tslib_module_info *input_mod_init(struct tsdev *dev, const char *params)
{
	struct tslib_input *i;

	tccdprint("input-raw init\r\n");
	
	i = malloc(sizeof(struct tslib_input));
	if (i == NULL)
		return NULL;

	i->module.ops = &__ts_input_ops;
	memset(i->current_x, 0, sizeof(i->current_x));
	memset(i->current_y, 0, sizeof(i->current_y));
	memset(i->current_p, 0, sizeof(i->current_p));
	memset(i->current_v, 0, sizeof(i->current_v));
	memset(i->current_pen_up, 0, sizeof(i->current_pen_up));
	
	i->sane_fd = -1;
	i->using_syn = 0;
	i->grab_events = 0;
	i->current_slot = 0;
	i->multi_touch = 0;
	i->multi_touch_need_fake = EMVARIANCE_FAKE_PRE;
	
	if (tslib_parse_vars(&i->module, raw_vars, NR_VARS, params)) {
		free(i);
		return NULL;
	}


	return &(i->module);
}

#ifndef TSLIB_STATIC_INPUT_MODULE
	TSLIB_MODULE_INIT(input_mod_init);
#endif 
